
__module_name__ = "_neural_diffeq.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# import packages #
# --------------- #
import torch


# import local dependencies #
# ------------------------- #
from ._compose_nn_sequential import _compose_nn_sequential


#### -------------------------------------------------------- ####


def _mu(mu, y):
    return mu(y)


def _sigma(sigma, y, brownian_size):
    return sigma(y).view(y.shape[0], y.shape[1], brownian_size)


#### -------------------------------------------------------- ####


class NeuralDiffEq(torch.nn.Module):

    noise_type = "general"
    sde_type = "ito"

    def __init__(self, mu, sigma=False, brownian_size=1):
        super(NeuralDiffEq, self).__init__()

        self.mu = mu
        self.sigma = sigma
        self._brownian_size = brownian_size

    def f(self, t, y0):
        return self.mu(y0)

    def forward(self, t, y0):
        return self.mu(y0)

    def g(self, t, y0):
        return _sigma(self.sigma, y0, self._brownian_size)
    
    
#### -------------------------------------------------------- ####
    
    
def _neural_diffeq(
    mu={1: [125, 125]},
    sigma={1: [125, 125]},
    in_dim=2,
    out_dim=2,
    mu_activation_function=torch.nn.Tanh(),
    sigma_activation_function=torch.nn.Tanh(),
    dropout=True,
    dropout_probability=0.5,
    brownian_size=1,
):
    """
    Instantiate a neural differential equation.

    Parameters:
    -----------
    in_dim
        Dimension size of input state
        type: int
        default: 2

    out_dim
        Dimension size of input state
        type: int
        default: 2

    mu
        dictionary-style composition of hidden architecture of the drift neural network
        type: dict
        default: {1: [125, 125]}

    sigma
        dictionary-style composition of hidden architecture of the diffusion neural network. Also may
        function as a boolean indicator to exclude a stochastic diffusion term. To negate stochastic
        term (thereby composing an ODE instead of an SDE), set `sigma=False`.
        type: dict (or bool)
        default: {1: [125, 125]}
    
    mu_activation_function
        Torch activation function of the mu (drift) neural network
        type: torch.nn.modules.activation
        default: torch.nn.Tanh()

    sigma_activation_function
        Torch activation function of the sigma (diffusion) neural network
        type: torch.nn.modules.activation
        default: torch.nn.Tanh()
    
    dropout
        Boolean indicator to include dropout in network architecture. If sigma is False, this
        argument is automatically adjusted to False.
        type: bool
        default: True

    dropout_probability
        Probability between 0 and 1 of node dropout for a given linear layer. If dropout is False,
        this argument is null.
        type: float
        default: 0.5

    brownian_size
        Dimension-wise complexity of the stochastic brownian noise.
        type: int
        default: 1

    Returns:
    --------
    NeuralDiffEq
    """

    if not sigma:
        dropout = False
        
    else:
        sigma = _compose_nn_sequential(
            in_dim=in_dim,
            out_dim=in_dim,
            hidden_layer_nodes=sigma,
            activation_function=sigma_activation_function,
            dropout=dropout,
            dropout_probability=dropout_probability,
        )

    mu = _compose_nn_sequential(
        in_dim=in_dim,
        out_dim=in_dim,
        hidden_layer_nodes=mu,
        activation_function=mu_activation_function,
        dropout=dropout,
        dropout_probability=dropout_probability,
    )

    return NeuralDiffEq(mu=mu, sigma=sigma, brownian_size=brownian_size)



__module_name__ = "_NeuralNetwork.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# import packages #
# --------------- #
import torch
from collections import OrderedDict


class _NeuralNetwork:
    def __init__(self):

        self._nn_dict = OrderedDict()
        self._activation_count = 0
        self._hidden_layer_count = 0
        self._dropout_count = 0

    def input_layer(self, in_dim, nodes):

        self._nn_dict["input_layer"] = torch.nn.Linear(in_dim, nodes)

    def activation_function(self, activation_function=torch.nn.LeakyReLU()):

        self._activation_count += 1
        self._nn_dict[
            "activation_{}".format(self._activation_count)
        ] = activation_function

    def dropout(self, probability=0.1):

        self._dropout_count += 1
        self._nn_dict["dropout_{}".format(self._dropout_count)] = torch.nn.Dropout(
            probability
        )

    def hidden_layer(self, nodes_m, nodes_n):

        self._hidden_layer_count += 1
        self._nn_dict["hidden_{}".format(self._hidden_layer_count)] = torch.nn.Linear(
            nodes_m, nodes_n
        )

    def output_layer(self, nodes, out_dim):

        self._nn_dict["output_layer"] = torch.nn.Linear(nodes, out_dim)

    def compose(self):

        return torch.nn.Sequential(self._nn_dict)



__module_name__ = "_compose_nn_sequential.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# import packages #
# --------------- #
import torch


# import local dependencies #
# ------------------------- #
from ._NeuralNetwork import _NeuralNetwork


def _compose_nn_sequential(
    in_dim=2,
    out_dim=2,
    activation_function=torch.nn.Tanh(),
    hidden_layer_nodes={1: [75, 75], 2: [75, 75]},
    dropout=True,
    dropout_probability=0.1,
):

    """Compose a sequential linear torch neural network"""

    nn = _NeuralNetwork()

    hidden_layer_keys = list(hidden_layer_nodes.keys())

    nn.input_layer(in_dim=in_dim, nodes=hidden_layer_nodes[hidden_layer_keys[0]][0])
    nn.activation_function(activation_function)

    for layer in hidden_layer_keys:
        layer_nodes = hidden_layer_nodes[layer]
        if dropout:
            nn.dropout(probability=dropout_probability)
        nn.hidden_layer(layer_nodes[0], layer_nodes[1])
        nn.activation_function(activation_function)

    if dropout:
        nn.dropout(probability=dropout_probability)
    nn.output_layer(out_dim=out_dim, nodes=hidden_layer_nodes[hidden_layer_keys[-1]][1])

    return nn.compose()